/*-
 * Copyright (c) 1999 Daniel Capo Sobral
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: conf.c,v 1.21 1999/12/19 21:33:03 dcs Exp $
 */


#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/stat.h>

#include "libconf.h"

#define UNREACHABLE_STATE	0
#define LINESINC		64

#define CMPVAR(conf,n1,n2)	(conf->casesensitive? \
				strcasecmp(n1,n2):strcmp(n1,n2))

/* General use structures */

struct configuration {		/* Configuration indexes */
	int		casesensitive;	/* Case sensitive? */
	struct conflvl*	levels;		/* List of configuration files */
	struct confsec*	sections;	/* List of sections */
	struct confsec* sectail;	/* Last section defined */
	struct confvar*	variables;	/* List of variables */
	struct confvar* vartail;	/* Last variable defined */
	struct confvardata* lastvar;	/* Last variable seen */
};

struct conflvl {		/* Level (file) information */
	int			level;		/* Configuration level */
	int			line;		/* First line with a change */
	int			changed;	/* Were any changes made? */
	char*			filename;	/* File name */
	char*			tmpfile;	/* Temporary name */
	char*			oldfile;	/* Hard link to old file */
	mode_t			mode;		/* File mode */
	uid_t			uid;		/* Owner id */
	gid_t			gid;		/* Group id */
	struct filecontent*	fc;		/* File contents' */
	struct conflvl*		next;		/* Next (higher) level */
	struct confvar*		variables;	/* List of variables */
	struct confsec*		sections;	/* List of sections */
	struct confsec*		sectail;	/* Last section defined */
	struct configuration*	conf;		/* Stop passing conf around */
	struct confinst*	instappended;	/* Appended instances */
	struct confinst*	instchanged;	/* Changed instances */
};

struct confsec {		/* Reference to a section */
	struct confsecdata*	data;		/* Section information */
	struct confsec*		next;		/* Next reference */
};

struct confsecdata {		/* Section information */
	struct conflvl*		level;		/* Configuration level */
	int			seclevel;	/* Section level */
	char*			name;		/* Section name */
	struct confdesc*	desc;		/* Section description */
	struct confvar*		variables;	/* Variables */
	struct confvar*		vartail;	/* Last variable defined */
	struct confsec*		subsections;	/* Subsections */
	struct confsec*		subsectail;	/* Last subsection defined */
	struct confsecdata*	parent;		/* Parent section */
};

struct confdesc {		/* Descriptions */
	long		nlines;	/* Number of lines */
	char**		line;	/* Description lines */
	char*		desc;	/* Single-string description (no linefeeds) */
};

struct filecontent {		/* File's content */
	char*			data;		/* Content */
	char*			endp;		/* End of data (+1) */
	long			nlines;		/* Number of lines */
	struct conflines**	lines;		/* Array of line indexes */
};

struct conflines {		/* Index to a file's lines */
	unsigned int		len;		/* Line length */
	long			line;		/* Pointer to line */
	struct confinst*	instance;	/* Variable instance */
};

struct confvar {		/* Generic list of variables */
	struct confvardata*	data;	/* Variable's data */
	struct confvar*		next;	/* Next variable */
};

struct confvardata {		/* General variable information */
	char*	name;		/* Name */
	struct confdesc* desc;	/* Description */
	char**	evalue;		/* Effective (highest level) value */
	char**	nvalue;		/* New value */
	int	llevel;		/* Lowest level at which it is defined */
	int	hlevel;		/* Highest level at which it is defined */
	int	nlevel;		/* Highest level at whcih it was changed */
	struct confinst*	instance;	/* List of instances */
	struct confinst*	insttail;	/* Last instance added */
	struct confinst*	cinstance;	/* List of changed instances */
};

struct confinst {		/* Generic list of variable instances */
	struct confinstdata*	data;	/* Instance's data */
	struct confinst*	next;	/* Next instance */
};

struct confinstdata {	/* Per instance variable information */
	struct conflvl*		level;	/* Level of definition */
	struct confsecdata*	section;/* Section of definition */
	long			line;	/* Line of definition */
	char*			value;	/* Original value */
	char*			nvalue;	/* New value */
	long			pvalue;	/* Pointer to the value in the line */
	struct confdesc*	desc;	/* Description */
	int		commented;	/* Commented-out definition? */
	struct confvardata*	data;	/* Variable information */
};

/* DFA enums */

enum trans { delim, tophead, subhead, assign, arem, vrem, bline };
enum states { start, thead, shead, vinst, hdeli, error };

#define NTRANS 7
#define NSTATES 6

/* Structures depending on dfa enums */

struct linebuf {
	char*		data;		/* Beginning of line */
	enum trans	type;		/* Type of line */
	int		nhash;		/* Number of hashes at the beginning */
	char*		ident;		/* Identifier */
	int		identsize;
	char*		text;		/* Name/Description/Value */
	unsigned int	textsize;
	char*		rem;		/* Comment */
	unsigned int	remsize;
};

/* DFA definitions */

static int noop();
static int newth(struct configuration*, struct linebuf*, int*);
static int newsh(struct configuration*, struct linebuf*, int*);
static int newssh(struct configuration*, struct linebuf*, int*);
static int newv(struct configuration*, struct linebuf*, int*);
static int athdesc(struct configuration*, struct linebuf*);
static int ashdesc(struct configuration*, struct linebuf*);
static int asshdesc(struct configuration*, struct linebuf*);
static int avdesc(struct configuration*, struct linebuf*);
static int endh(struct configuration*, struct linebuf*);
static int endv(struct configuration*, struct linebuf*);
static int adjsh(struct configuration*, struct linebuf*, int*);
static int adjssh(struct configuration*, struct linebuf*, int*);

static struct trtbl_s {
	enum states newstate;
	int (*preaction)(struct configuration*, struct linebuf*);
	int (*posaction)(struct configuration*, struct linebuf*, int*);
	} trtbl[NTRANS][NSTATES] = {
	{ /* Header delimiter line */ 
		{ hdeli, noop, noop }, { start, endh, noop },
		{ hdeli, endh, noop }, { hdeli, endv, noop },
		{ hdeli, noop, noop }, { start, noop, noop } },
	{ /* Top header line */
		{ shead, noop,    newsh }, { thead, athdesc, noop },
		{ shead, ashdesc, adjsh }, { shead, endv,    newth },
		{ thead, noop,    newth }, { shead, noop,    newth } },
	{ /* Sub header line */
		{ shead, noop,     newssh }, { error, endh, noop   },
		{ shead, asshdesc, adjssh }, { shead, endv, newssh },
		{ shead, noop,     newssh }, { shead, noop, newssh } },
	{ /* Assignment */
		{ vinst, noop, newv}, { error, endh, noop },
		{ vinst, endh, newv}, { vinst, endv, newv },
		{ vinst, noop, newv}, { vinst, noop, newv } },
	{ /* Commented out variable assignment */
		{ start, noop, noop }, { error, endh, noop },
		{ start, endh, noop }, { start, endv, noop },
		{ start, noop, noop }, { start, noop, noop } },
	{ /* Variable description */
		{ start, noop, noop }, { error, endh,   noop },
		{ start, endh, noop }, { vinst, avdesc, noop },
		{ start, noop, noop }, { start, noop, noop } },
	{ /* Any other type of line */
		{ start, noop, noop }, { start, endh, noop },
		{ start, endh, noop }, { start, endv, noop },
		{ start, noop, noop }, { start, noop, noop } }
};

/* Function prototypes */

static int delsection(struct configuration*);
static int parseline(struct linebuf*, char**);
static int parsefile(struct configuration*, const char*);
static int addsection(struct configuration*, struct confsecdata*, int);
static int addinstance(struct configuration*, struct confinstdata*, const char*, unsigned int);
static int addvardesc(struct confinstdata*, struct linebuf*);
static int addsecdesc(struct confsecdata*, struct linebuf*);
static int adddesc(struct confdesc*, const char*, unsigned int);
static int addline(struct configuration*, struct confinstdata*);
static struct confvardata* getvariable(struct configuration*, struct conflvl*, const char*);
static struct confvardata* createvariable(const char*);
static struct confvar* createvarref(struct confvardata*);
static struct confsec* createsecref(struct confsecdata*);
static void addinstchange(struct configuration*, struct confinstdata*);
static void removeinstchange(struct configuration*, struct confinstdata*);
static void freedesc(struct confdesc*);
static void adjustevalue(struct configuration*);

static int
parseline(struct linebuf* line, char** p)
{
	int	s;

	line->data = *p;
	line->nhash = 0;
	line->identsize = 0;
	line->textsize = 0;
	line->remsize = 0;

	s = 0;

	if(**p == NULL)
		return 0;

	while(**p && **p != '\n') {
		switch(s) {
		    case 0: /* Beginning of line */
			if(**p == '#') {
				line->nhash++;
				while(*++*p == '#')
					line->nhash++;
				s = 1;
				break;
			}
			if(isalpha(**p) || **p == '_') {
				line->ident = line->text = *p;
				line->identsize = 1;
				(*p)++;
				while(isalnum(**p) || **p == '.' || **p == '_') {
					line->identsize++;
					(*p)++;
				}
				s = 2;
				break;
			}
			if(isspace(**p)) {
				do
					(*p)++;
				while(isspace(**p) && **p != '\n');
				s = 3;
				break;
			}
			do
				(*p)++;
			while(**p && **p != '\n');
			s = -1;
			break;
		    case 1: /* #'s at the beginning of the line */
			if(line->nhash == 1 && (isalpha(**p) || **p == '_')) {
				line->ident = line->text = *p;
				line->identsize = 1;
				line->textsize = 1;
				(*p)++;
				while(isalnum(**p) || **p == '.' || **p == '_') {
					line->identsize++;
					line->textsize++;
					(*p)++;
				}
				s = 4;
				break;
			}
			if(isspace(**p)) {
				do
					(*p)++;
				while(isspace(**p) && **p != '\n');
				if(line->nhash == 1)
					s = 5;
				else
					s = 12;
				break;
			}
			line->text = *p;
			line->textsize = 1;
			while(*++*p && **p != '#' && **p != '\n')
				line->textsize++;
			s = 6;
			break;
		    case 2: /* Possible assignment */
			if(isspace(**p)) {
				do
					(*p)++;
				while(isspace(**p) && **p != '\n');
				s = 7;
				break;
			}
			if(**p == '=') {
				(*p)++;
				s = 8;
				break;
			}
			do
				(*p)++;
			while(**p && **p != '\n');
			s = -1;
			break;
		    case 3: /* Possible variable comment */
			if(**p == '#') {
				do
					(*p)++;
				while(**p == '#');
				while(isspace(**p) && **p != '\n')
					(*p)++;
				line->rem = *p;
				line->remsize = 0;
				while(**p && **p != '\n') {
					line->remsize++;
					(*p)++;
				}
				s = 9;
				break;
			}
			do
				(*p)++;
			while(**p && **p != '\n');
			s = -1;
			break;
		    case 4: /* Possible commented assignment */
			if(isspace(**p)) {
				do {
					(*p)++;
					line->textsize++;
				} while(isspace(**p) && **p != '\n');
				s = 10;
				break;
			}
			if(**p == '=') {
				(*p)++;
				line->textsize++;
				s = 11;
				break;
			}
			while(*++*p && **p != '#' && **p != '\n')
				line->textsize++;
			s = 6;
			break;
		    case 5: /* #<space><something> ... */
			if(isalpha(**p) || **p == '_') {
				line->text = *p;
				line->textsize = 1;
				(*p)++;
				while(isalnum(**p) || **p == '.' || **p == '_') {
					line->textsize++;
					(*p)++;
				}
				s = 13;
				break;
			}
			line->text = *p;
			line->textsize = 1;
			while(*++*p && **p != '#' && **p != '\n')
				line->textsize++;
			s = 6;
			break;
		    case 6: /* If it got here at all, it's a top header */
			do
				(*p)++;
			while(**p && **p != '\n');
			s = 14;
			break;
		    case 7: /* Possible assignment */
			if(**p == '=') {
				(*p)++;
				s = 8;
				break;
			}
			do
				(*p)++;
			while(**p && **p != '\n');
			s = -1;
			break;
		    case 8: /* Possible assignment */
			if(isspace(**p)) {
				do
					(*p)++;
				while(isspace(**p) && **p != '\n');
			}
			if(**p == '"') {
				line->text = (*p)++;
				line->textsize = 1;
				if(**p == '\\') {
					(*p)++;
					line->textsize++;
				}
				while(**p && **p != '\n' && **p != '"') {
					(*p)++;
					line->textsize++;
					if(**p == '\\') {
						(*p)++;
						line->textsize++;
					}
				}
				if(**p == '"') {
					s = 15;
					(*p)++;
					line->textsize++;
					while(isspace(**p) && **p != '\n')
						(*p)++;
				} else
					s = 16;
				break;
			}
			if(isalnum(**p) || **p == '_') {
				line->text = (*p)++;
				line->textsize = 1;
				while(isalnum(**p) || **p == '_' ||
				    **p == '.') {
					(*p)++;
					line->textsize++;
				}
				if(isspace(**p) || **p == '#' || **p == '\n'
				    || !**p) {
					s = 15;
					while(isspace(**p) && **p != '\n')
						(*p)++;
				} else
					s = 16;
				break;
			}
			do
				(*p)++;
			while(**p && **p != '\n');
			s = -1;
			break;
		    case 9: /* NOT REACHED -- variable comment */
			s = -1;
			break;
		    case 10: /* Possible commented assignment */
			if(**p == '=') {
				(*p)++;
				line->textsize++;
				s = 11;
				break;
			}
			while(*++*p && **p != '#' && **p != '\n')
				line->textsize++;
			s = 6;
			break;
		    case 11: /* Possible commented assignment */
			if(isspace(**p)) {
				do {
					(*p)++;
					line->textsize++;
				} while(isspace(**p) && **p != '\n');
			}
			if(**p == '"') {
				line->text = (*p)++;
				line->textsize = 1;
				if(**p == '\\') {
					(*p)++;
					line->textsize++;
				}
				while(**p && **p != '\n' && **p != '"') {
					(*p)++;
					line->textsize++;
					if(**p == '\\') {
						(*p)++;
						line->textsize++;
					}
				}
				if(**p == '"') {
					s = 15;
					(*p)++;
					line->textsize++;
					while(isspace(**p) && **p != '\n')
						(*p)++;
				} else
					s = 16;
				break;
			}
			if(isalnum(**p) || **p == '_') {
				line->text = (*p)++;
				line->textsize = 1;
				while(isalnum(**p) || **p == '_' ||
				    **p == '.') {
					(*p)++;
					line->textsize++;
				}
				if(isspace(**p) || **p == '#' || **p == '\n'
				    || !**p) {
					s = 15;
					while(isspace(**p) && **p != '\n')
						(*p)++;
				} else
					s = 16;
				break;
			}
			while(**p && **p != '#' && **p != '\n') {
				(*p)++;
				line->textsize++;
			}
			s = 6;
			break;
		    case 12: /* Header */
			line->text = *p;
			line->textsize = 1;
			while(*++*p && **p != '#' && **p != '\n')
				line->textsize++;
			s = 6;
			break;
		    case 13: /* Possible truly commented assignment */
			if(isspace(**p)) {
				do {
					line->textsize++;
					(*p)++;
				} while(isspace(**p) && **p != '\n');
			}
			if(**p == '=') {
				line->textsize++;
				(*p)++;
				s = 18;
				break;
			}
			while(**p && **p != '#' && **p != '\n') {
				line->textsize++;
				(*p)++;
			}
			s = 6;
			break;
		    case 14: /* NOT REACHED -- top header */
			s = -1;
			break;
		    case 15: /* Assignment */
			if(**p == '#') {
				do
					(*p)++;
				while(**p == '#');
				while(isspace(**p) && **p != '\n')
					(*p)++;
				line->rem = *p;
				line->remsize = 0;
				while(**p && **p != '\n') {
					line->remsize++;
					(*p)++;
				}
				s = 19;
				break;
			}
			do
				(*p)++;
			while(**p && **p != '\n');
			s = 20;
			break;
		    case 16: /* Assignment with syntax error */
			do
				(*p)++;
			while(**p && **p != '\n');
			break;
		    case 17: /* UNREACHABLE */
			s = -1;
			break;
		    case 18: /* Possible truly commented assignment */
			if(isspace(**p)) {
				do {
					(*p)++;
					line->textsize++;
				} while(isspace(**p) && **p != '\n');
			}
			if(**p == '"') {
				do
					(*p)++;
				while(**p && **p != '\n');
				s = 21;
				break;
			}
			if(isalnum(**p) || **p == '_') {
				do
					(*p)++;
				while(**p && **p != '\n');
				s = 21;
				break;
			}
			while(**p && **p != '#' && **p != '\n') {
				(*p)++;
				line->textsize++;
			}
			s = 6;
			break;
		    case 19: /* NOT REACHED -- assignment */
			s = -1;
			break;
		    case 20: /* NOT REACHED -- assignment with syntax error */
			s = -1;
			break;
		    case 21: /* NOT REACHED -- truly commented assignment */
			s = -1;
			break;
		    case -1: /* NOT REACHED -- stuff */
			assert(UNREACHABLE_STATE);
			break;
		    default: /* NOT REACHED */
			s = -1;
			break;
		}
	}

	if(**p == '\n')
		(*p)++;

	switch(s) {
	    case 0:
		line->type = bline;
		return 1;
	    case 1:
		line->type = delim;
		return 1;
	    case 2:
		line->type = bline;
		return 1;
	    case 3:
		line->type = bline;
		return 1;
	    case 4:
		line->type = subhead;
		return 1;
	    case 5:
		line->type = delim;
		return 1;
	    case 6:
		line->type = subhead;
		return 1;
	    case 7:
		line->type = bline;
		return 1;
	    case 8:
		line->type = bline;
		return 1;
	    case 9:
		line->type = vrem;
		return 1;
	    case 10:
		line->type = subhead;
		return 1;
	    case 11:
		line->type = subhead;
		return 1;
	    case 12:
		line->type = subhead;
		return 1;
	    case 13:
		line->type = subhead;
		return 1;
	    case 14:
		line->type = tophead;
		return 1;
	    case 15:
		line->type = assign;
		return 1;
	    case 16:
		line->type = assign;
		return -1;
	    case 17:
		line->type = subhead;
		return 1;
	    case 18:
		line->type = subhead;
		return 1;
	    case 19:
		line->type = assign;
		return 1;
	    case 20:
		line->type = assign;
		return -1;
	    case 21:
		line->type = bline;
		return 1;
	    case -1:
		line->type = bline;
		return 1;
	    default:
		assert(UNREACHABLE_STATE);
	}
	return 0;
}

static int
parsefile(struct configuration* conf, const char* filename)
{
	struct filecontent*	fc;
	struct linebuf		line;
	struct stat		sb;
	off_t			filesize;
	int			fd;
	int			state, newstate;
	int			errorcode, rvalue;
	char			*p;
	
	fc = conf->levels->fc;
	rvalue = 0;

	/* Read file */

	if((fd = open(filename, O_RDONLY)) == -1)
		return ENOENT;

	if((filesize = lseek(fd, (off_t)0, SEEK_END)) == -1)
		return errno;
	if(lseek(fd, (off_t)0, SEEK_SET) == -1)
		return errno;

	if((fc->data = (char*) malloc((size_t)filesize+1)) == NULL)
		return ENOMEM;

	if(read(fd, fc->data, (size_t)filesize) == -1) {
		free(fc->data);
		fc->data = NULL;
		return errno;
	}

	fc->endp = fc->data + filesize;
	*fc->endp = 0;

	if(fstat(fd, &sb) == -1) {
		conf->levels->uid = 0;
		conf->levels->gid = 0;
		conf->levels->mode = 0;
		return CANT_STAT;
	} else {
		conf->levels->uid = sb.st_uid;
		conf->levels->gid = sb.st_gid;
		conf->levels->mode = sb.st_mode;
	}

	while(close(fd) == -1)
		if(errno != EINTR) {
			free(fc->data);
			fc->data = NULL;
			return errno;
		}

	/* Parse file */

	p = fc->data;

	state = start;

	while((errorcode = parseline(&line, &p))) {
		if(errorcode == -1)
			rvalue = SYNTAX_ERROR;

		if(fc->nlines % LINESINC == 0)
			if((fc->lines = (struct conflines**)
			    reallocf(fc->lines,
			    (fc->nlines / LINESINC + 1) *
			    LINESINC * sizeof(struct conflines*))) == NULL)
				return ENOMEM;

		fc->lines[fc->nlines] = (struct conflines*)
		    malloc(sizeof(struct conflines));
		fc->lines[fc->nlines]->line = line.data - fc->data;
		fc->lines[fc->nlines]->len = strcspn(line.data, "\n");
		fc->lines[fc->nlines]->instance = NULL;
		fc->nlines++;
		if((errorcode = trtbl[line.type][state].preaction(conf, &line))
		    != 0)
			 rvalue = errorcode;
		newstate = trtbl[line.type][state].newstate;
		trtbl[line.type][state].posaction(conf, &line, &newstate);
		state = newstate;
	}

	return rvalue;
}

static int
noop()
{
	return 0;
}

static int
newth(struct configuration* conf, struct linebuf* line, int* state)
{
	struct confsecdata*	sec;

	delsection(conf);

	sec = (struct confsecdata*) malloc(sizeof(struct confsecdata));

	sec->level = conf->levels;
	sec->seclevel = 0;
	sec->name = (char*) malloc(line->textsize+1);
	strncpy(sec->name, line->text, line->textsize);
	sec->name[line->textsize] = '\0';
	sec->desc = (struct confdesc*) malloc(sizeof(struct confdesc));
	sec->desc->nlines = 0;
	sec->desc->line = NULL;
	sec->desc->desc = NULL;
	sec->variables = NULL;
	sec->subsections = NULL;
	sec->parent = NULL;

	addsection(conf, sec, 0);

	return 0;
}

static int
newsh(struct configuration* conf, struct linebuf* line, int* state)
{
	struct confsecdata*	sec;

	if(conf->sections && conf->sectail->data->seclevel >= 1)
		delsection(conf);

	sec = (struct confsecdata*) malloc(sizeof(struct confsecdata));

	sec->level = conf->levels;
	sec->seclevel = 1;
	sec->name = (char*) malloc(line->textsize+1);
	strncpy(sec->name, line->text, line->textsize);
	sec->name[line->textsize] = '\0';
	sec->desc = (struct confdesc*) malloc(sizeof(struct confdesc));
	sec->desc->nlines = 0;
	sec->desc->line = NULL;
	sec->desc->desc = NULL;
	sec->variables = NULL;
	sec->subsections = NULL;

	addsection(conf, sec, 1);

	return 0;
}

static int
newssh(struct configuration* conf, struct linebuf* line, int* state)
{
	struct confsecdata*	sec;

	if(conf->sections && conf->sectail->data->seclevel >= 2)
		delsection(conf);

	sec = (struct confsecdata*) malloc(sizeof(struct confsecdata));

	sec->level = conf->levels;
	sec->seclevel = 2;
	sec->name = (char*) malloc(line->textsize+1);
	strncpy(sec->name, line->text, line->textsize);
	sec->name[line->textsize] = '\0';
	sec->desc = (struct confdesc*) malloc(sizeof(struct confdesc));
	sec->desc->nlines = 0;
	sec->desc->line = NULL;
	sec->desc->desc = NULL;
	sec->variables = NULL;
	sec->subsections = NULL;
	sec->parent = NULL;

	addsection(conf, sec, 2);

	return 0;
}

static int
newv(struct configuration* conf, struct linebuf* line, int* state)
{
	struct confinstdata*	varinst;

	varinst = (struct confinstdata*) malloc(sizeof(struct confinstdata));
	varinst->level = conf->levels;
	if(conf->levels->sections)
		varinst->section = conf->levels->sectail->data;
	else
		varinst->section = NULL;
	varinst->line = conf->levels->fc->nlines-1;
	if(*line->text == '"' || *line->text == '\'') {
		varinst->value = (char*) malloc(line->textsize-1);
		strncpy(varinst->value, line->text+1, line->textsize-2);
		varinst->value[line->textsize-2] = '\0';
	} else {
		varinst->value = (char*) malloc(line->textsize+1);
		strncpy(varinst->value, line->text, line->textsize);
		varinst->value[line->textsize] = '\0';
	}
	varinst->nvalue = NULL;
	varinst->pvalue = line->text - conf->levels->fc->data;
	varinst->desc = (struct confdesc*) malloc(sizeof(struct confdesc));
	varinst->desc->nlines = 0;
	varinst->desc->line = NULL;
	varinst->desc->desc = NULL;
	addvardesc(varinst, line);
	varinst->commented = line->nhash > 0;

	addinstance(conf, varinst, line->ident, line->identsize);
	addline(conf, varinst);

	return 0;
}

static int
endh(struct configuration* conf, struct linebuf* line)
{
	return 0;
}

static int
endv(struct configuration* conf, struct linebuf* line)
{
	return 0;
}

static int
athdesc(struct configuration* conf, struct linebuf* line)
{
	addsecdesc(conf->levels->sectail->data, line);

	return 0;
}

static int
ashdesc(struct configuration* conf, struct linebuf* line)
{
	int	errorcode;

	if(1 != conf->levels->sectail->data->seclevel) {
		errorcode = endh(conf, line);
		delsection(conf);
		return errorcode;
	}

	addsecdesc(conf->levels->sectail->data, line);

	return 0;
}

static int
asshdesc(struct configuration* conf, struct linebuf* line)
{
	int	errorcode;

	if(2 != conf->levels->sectail->data->seclevel) {
		errorcode = endh(conf, line);
		delsection(conf);
		return errorcode;
	}

	addsecdesc(conf->levels->sectail->data, line);

	return 0;
}

static int
avdesc(struct configuration* conf, struct linebuf* line)
{
	addvardesc(conf->lastvar->insttail->data, line);

	return 0;
}

static int
addsection(struct configuration* conf, struct confsecdata* sec, int seclevel)
{
	struct confsec*		ref;
	struct confsecdata*	p;

	/* List of all sections */
	ref = createsecref(sec);
	if(conf->sections == NULL)
		conf->sections = ref;
	else
		conf->sectail->next = ref;
	conf->sectail = ref;

	/* List of subsections to a section */
	if(conf->levels->sections != NULL) {
		p = conf->levels->sectail->data;
		while(p && p->seclevel >= seclevel )
			p = p->parent;
		if(p) {
			sec->parent = p;
			ref = createsecref(sec);
			if(p->subsections == NULL)
				p->subsections = ref;
			else
				p->subsectail->next = ref;
			p->subsectail = ref;
		}
	}

	/* List of all sections in a configuration level */	
	ref = createsecref(sec);
	if(conf->levels->sections == NULL)
		conf->levels->sections = ref;
	else
		conf->levels->sectail->next = ref;
	conf->levels->sectail = ref;

	return 0;
}

static int
addinstance(struct configuration* conf, struct confinstdata* instdata, const char* name, unsigned int len)
{
	struct confvardata*	var;
	struct confinst*	inst;
	char*			varname;

	varname = (char*) malloc(len+1);
	strncpy(varname, name, len);
	varname[len] = '\0';

	var = getvariable(conf, conf->levels, varname);
	conf->lastvar = var;
	instdata->data = var;

	free(varname);

	inst = (struct confinst*) malloc(sizeof(struct confinst));
	inst->data = instdata;
	inst->next = NULL;

	if(var->instance == NULL)
		var->instance = inst;
	else
		var->insttail->next = inst;
	var->insttail = inst;

	if(var->desc == NULL)
		var->desc = instdata->desc;

	if(!instdata->commented) {
		if(var->llevel == -1)
			var->llevel = conf->levels->level;
		var->hlevel = conf->levels->level;
		var->evalue = &instdata->value;
		var->nvalue = &instdata->nvalue;
	}

	return 0;
}

static struct confvardata*
getvariable(struct configuration* conf, struct conflvl* level, const char* name)
{
	struct confvardata*	var;
	struct confvar*		ref;
	struct confvar*		p;

	if(conf->variables == NULL) {
		var = createvariable(name);
		ref = createvarref(var);
		conf->variables = ref;
		conf->vartail = ref;
		ref->next = NULL;
	} else {
		p = conf->variables;

		if(CMPVAR(conf, p->data->name, name) > 0) {
			var = createvariable(name);
			ref = createvarref(var);
			conf->variables = ref;
			conf->vartail = ref;
			conf->variables->next = p;
		} else {
			while(p->next &&
			    CMPVAR(conf, p->next->data->name, name) <= 0)
				p = p->next;

			if(CMPVAR(conf, p->data->name, name) == 0) {
				var = p->data;
			} else {
				if(p->next == NULL) {
					var = createvariable(name);
					ref = createvarref(var);
					p->next = ref;
					conf->vartail = ref;
					p->next->next = NULL;
				} else {
					var = createvariable(name);
					ref = createvarref(var);
					ref->next = p->next;
					conf->vartail = ref;
					p->next = ref;
				}
			}
		}
	}

	if(level->variables == NULL) {
		ref = createvarref(var);
		level->variables = ref;
		ref->next = NULL;
	} else {
		p = level->variables;

		if(CMPVAR(conf, p->data->name, name) > 0) {
			level->variables = createvarref(var);
			level->variables->next = p;
		} else {
			while(p->next &&
			    CMPVAR(conf, p->next->data->name, name) <= 0)
				p = p->next;

			if(CMPVAR(conf, p->data->name, name) != 0) {
				if(p->next == NULL) {
					p->next = createvarref(var);
					p->next->next = NULL;
				} else {
					ref = createvarref(var);
					ref->next = p->next;
					p->next = ref;
				}
			}
		}
	}

	if(level->sections != NULL) {
		if(level->sectail->data->variables == NULL) {
			ref = createvarref(var);
			level->sectail->data->variables = ref;
			level->sectail->data->vartail = ref;
			ref->next = NULL;
		} else {
			p = level->sectail->data->variables;

			while(p && CMPVAR(conf, p->data->name, name) != 0)
				p = p->next;

			if(p == NULL) {
				ref = createvarref(var);
				level->sectail->data->vartail->next
				    = ref;
				level->sectail->data->vartail = ref;
				ref->next = NULL;
			}
		}
	}

	return var;
}

static struct confvardata*
createvariable(const char* name)
{
	struct confvardata*	var;

	var = (struct confvardata*) malloc(sizeof(struct confvardata));

	var->name = strdup(name);
	var->desc = NULL;
	var->evalue = NULL;
	var->nvalue = NULL;
	var->llevel = -1;
	var->hlevel = -1;
	var->nlevel = -1;
	var->instance = NULL;
	var->insttail = NULL;
	var->cinstance = NULL;

	return var;
}

static struct confsec*
createsecref(struct confsecdata* sec)
{
	struct confsec*	ref;

	ref = (struct confsec*) malloc(sizeof(struct confsec));
	ref->data = sec;
	ref->next = NULL;

	return ref;
}

static struct confvar*
createvarref(struct confvardata* var)
{
	struct confvar*	ref;

	ref = (struct confvar*) malloc(sizeof(struct confvar));
	ref->data = var;
	ref->next = NULL;

	return ref;
}


static int
addvardesc(struct confinstdata* var, struct linebuf* line)
{
	return adddesc(var->desc, line->rem, line->remsize);
}

static int
addsecdesc(struct confsecdata* sec, struct linebuf* line)
{
	return adddesc(sec->desc, line->text, line->textsize);
}

static int
adddesc(struct confdesc* desc, const char* text, unsigned int len)
{
	if(len == 0)
		return 0;

	desc->line = (char**) realloc(desc->line,
	    (desc->nlines + 1) * sizeof(char*));
	desc->line[desc->nlines] = (char*) malloc(len + 1);
	strncpy(desc->line[desc->nlines], text, len);
	desc->line[desc->nlines][len] = '\0';
	if(desc->desc == NULL)
		desc->desc = strdup(desc->line[desc->nlines]);
	else {
		desc->desc = (char*) realloc(desc->desc, strlen(desc->desc) +
		    len + 2);
		strcat(desc->desc, " ");
		strncat(desc->desc, text, len);
	}
	desc->nlines++;

	return 0;
}

static int
adjsh(struct configuration* conf, struct linebuf* line, int* state)
{
	int	errorcode;

	if(conf->sectail->data->seclevel == 1)
		return 0;

	errorcode = newsh(conf, line, state);
	*state = shead;

	return errorcode;
}

static int
adjssh(struct configuration* conf, struct linebuf* line, int* state)
{
	int	errorcode;

	if(conf->sectail->data->seclevel == 2)
		return 0;

	errorcode = newsh(conf, line, state);
	*state = shead;

	return errorcode;
}

static int
delsection(struct configuration* conf)
{
	struct confsec*		ref;	/* iterator */
	struct confsec*		fref;	/* reference to be freed */
	struct confsecdata*	p;	/* iterator */
	struct confsecdata*	sec;	/* section being removed */

	if(conf->sections == NULL)
		return 0;

	sec = conf->sectail->data;

	if(sec->level != conf->levels)
		return 0;

	if(sec->variables != NULL || sec->subsections != NULL)
		return 0;

	if((p = sec->parent) != NULL) {
		ref = p->subsections;
		if(ref->data == sec) {
			p->subsections = NULL;
			p->subsectail = NULL;
			fref = ref;
		} else {
			while(ref->next->data != sec)
				ref = ref->next;
			fref = ref->next;
			ref->next = ref->next->next;
			if(ref->next == NULL)
				p->subsectail = ref;
		}
		free(fref);
	}

	ref = conf->levels->sections;
	if(ref->data == sec) {
		conf->levels->sections = NULL;
		conf->levels->sectail = NULL;
		fref = ref;
	} else {
		while(ref->next->data != sec)
			ref = ref->next;
		fref = ref->next;
		ref->next = ref->next->next;
		if(ref->next == NULL)
			conf->levels->sectail = ref;
	}
	free(fref);

	ref = conf->sections;
	if(ref->data == sec) {
		conf->sections = NULL;
		conf->sectail = NULL;
		fref = ref;
	} else {
		while(ref->next->data != sec)
			ref = ref->next;
		fref = ref->next;
		ref->next = ref->next->next;
		if(ref->next == NULL)
			conf->sectail = ref;
	}
	free(fref);

	return 0;
}

static int
addline(struct configuration* conf, struct confinstdata* var)
{
	struct confinst*	inst;

	inst = (struct confinst*) malloc(sizeof(struct confinst));
	inst->data = var;
	inst->next = NULL;

	conf->levels->fc->lines[conf->levels->fc->nlines-1]->instance = inst;

	return 0;
}

/* Exported functions */

struct configuration*
createconf(int casesensitive)
{
	struct configuration*	conf;

	conf = (struct configuration*) malloc(sizeof(struct configuration));

	conf->casesensitive = casesensitive;
	conf->levels = NULL;
	conf->sections = NULL;
	conf->sectail = NULL;
	conf->variables = NULL;
	conf->vartail = NULL;
	conf->lastvar = NULL;

	return conf;
}

int
readconf(struct configuration* conf, const char* filename)
{
	struct conflvl*		level;
	struct filecontent*	fc;
	int			errorcode;

	level = (struct conflvl*) malloc(sizeof(struct conflvl));
	fc = (struct filecontent*) malloc(sizeof(struct filecontent));

	if(conf->levels)
		level->level = conf->levels->level + 1;
	else
		level->level = 0;
	level->line = -1;
	level->changed = 0;
	level->filename = strdup(filename);
	level->tmpfile = NULL;
	level->oldfile = NULL;
	level->fc = fc;
	level->instappended = NULL;
	level->instchanged = NULL;
	level->variables = NULL;
	level->sections = NULL;
	level->sectail = NULL;

	fc->data = NULL;
	fc->endp = NULL;
	fc->nlines = 0;
	fc->lines = NULL;

	level->next = conf->levels;
	conf->levels = level;
	level->conf = conf;

	errorcode = parsefile(conf, filename);

	return errorcode;
}

struct confvar*
getfirstvar(struct configuration* conf)
{
	return conf->variables;
}

struct confvar*
getfirstvarlevel(struct configuration* conf, int level)
{
	struct conflvl*	p;

	p = conf->levels;

	while(p && p->level != level)
		p = p->next;

	if(p != NULL)
		return p->variables;
	else
		return NULL;
}

struct confvar*
getfirstvarsec(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->data->variables;
}

struct confsec*
getfirstsection(struct configuration* conf)
{
	return conf->sections;
}

struct confsec*
getfirstsectionlevel(struct configuration* conf, int level)
{
	struct conflvl*	p;

	p = conf->levels;

	while(p && p->level != level)
		p = p->next;

	if(p != NULL)
		return p->sections;
	else
		return NULL;
}

struct confsec*
getfirstsubsection(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->data->subsections;
}

struct confvar*
getnextvar(struct confvar* var)
{
	if(var == NULL)
		return NULL;
	return var->next;
}

struct confsec*
getnextsec(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->next;
}

const char*
getvardesc(struct confvar* var)
{
	if(var == NULL)
		return NULL;
	return var->data->desc->desc;
}

const char*
getvardesclevel(struct confvar* var, int level)
{
	struct confinst*	p;

	p = var->data->instance;

	while(p && p->data->level->level != level)
		p = p->next;

	if(p != NULL)
		return p->data->desc->desc;
	else
		return NULL;
}

const char*
getsecdesc(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->data->desc->desc;
}

int
getsecdescnlines(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->data->desc->nlines;
}

const char*
getsecdescline(struct confsec* sec, int line)
{
	if(sec == NULL)
		return NULL;
	return sec->data->desc->line[line];
}

const char*
getsecname(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->data->name;
}

const char*
getvarname(struct confvar* var)
{
	if(var == NULL)
		return NULL;
	return var->data->name;
}

const char*
getvarvalue(struct confvar* var)
{
	if(var == NULL)
		return NULL;
	if(var->data->evalue == NULL)
		return NULL;
	if(*var->data->nvalue != NULL && var->data->nlevel >= var->data->hlevel)
		return *var->data->nvalue;
	else
		return *var->data->evalue;
}

const char*
getvarovalue(struct confvar* var)
{
	if(var == NULL)
		return NULL;
	return *var->data->evalue;
}

const char*
getvarvaluelevel(struct confvar* var, int level)
{
	struct confinst*	p;

	p = var->data->instance;

	while(p && p->data->level->level != level)
		p = p->next;

	while(p && p->next && p->next->data->level->level == level)
		p = p->next;

	if(p != NULL) {
		if(p->data->nvalue == NULL)
			return p->data->value;
		else
			return p->data->nvalue;
	} else
		return NULL;
}

int
getvarlevel(struct confvar* var)
{
	if(var == NULL)
		return NULL;
	return var->data->hlevel;
}

int
getseclevel(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->data->seclevel;
}

int
getsecconflevel(struct confsec* sec)
{
	if(sec == NULL)
		return NULL;
	return sec->data->level->level;
}

int
confchanged(struct configuration* conf)
{
	struct conflvl*	p;

	p = conf->levels;
	while(p) {
		if(p->changed)
			return 1;
		p = p->next;
	}

	return 0;
}

int
confchangedlevel(struct configuration* conf, int level)
{
	struct conflvl*	p;

	p = conf->levels;

	while(p && p->level != level)
		p = p->next;

	if(p != NULL)
		return p->changed;
	else
		return 0;
}

int
varchanged(struct confvar* var)
{
	return var->data->cinstance != NULL;
}

struct confsec*
getvarsec(struct confvar* var)
{
	struct confsec*	p;

	p = var->data->instance->data->level->conf->sections;

	while(p && p->data != var->data->instance->data->section)
		p = p->next;

	return p;
}

struct confsec*
getvarseclevel(struct confvar* var)
{
	struct confsec*	p;

	p = var->data->instance->data->level->sections;

	while(p && p->data != var->data->instance->data->section)
		p = p->next;

	return p;
}

struct confsec*
getsecsec(struct confsec* sec)
{
	struct confsec*		p;
	struct confsecdata*	parent;

	if(sec->data->parent ==  NULL)
		return NULL;

	parent = sec->data->parent;

	if(parent->parent == NULL)
		p = sec->data->level->sections;
	else
		p = parent->parent->subsections;

	while(p && p->data != parent)
		p = p->next;

	return p;
}

int
varcommented(struct confvar* var, int level)
{
	struct confinst*	p;
	struct confinst*	aux;

	p = var->data->instance;

	while(p && p->data->level->level != level)
		p = p->next;

	aux = p;
	while(aux && aux->next && aux->next->data->level->level == level) {
		aux = aux->next;
		if(!aux->data->commented)
			p = aux;
	}

	if(p != NULL)
		return p->data->commented;
	else
		return NULL;
}

int
setvarlevel(struct confvar* var, int level, const char* value)
{
	struct configuration*	conf;
	struct confinst*	p;
	struct confinst*	aux;

	conf = var->data->instance->data->level->conf;

	if(level > conf->levels->level)
		return -1;

	p = var->data->instance;
	while(p && (p->data->level->level != level || p->data->commented))
		p = p->next;

	if(p == NULL)
		p = appendvarinstlevel(conf, level, var->data->name);
	else {
		aux = p;
		while(aux->next && aux->next->data->level->level == level) {
			aux = aux->next;
			if(!aux->data->commented)
				p = aux;
		}
	}

	if(var->data->nlevel <= level) {
		var->data->nvalue = &p->data->nvalue;
		var->data->nlevel = level;
	}

	if(p->data->nvalue != NULL)
		free(p->data->nvalue);
	else
		addinstchange(conf, p->data);
	p->data->nvalue = strdup(value);

	p->data->level->changed = 1;

	return 0;
}

int
unsetvarlevel(struct confvar* var, int level)
{
	struct configuration*	conf;
	struct confinst*	p;
	struct confinst*	aux;

	conf = var->data->instance->data->level->conf;

	if(level > conf->levels->level)
		return -1;

	if(var->data->nvalue == NULL || *var->data->nvalue == NULL)
		return 0;

	p = var->data->instance;
	while(p && (p->data->level->level != level || p->data->commented))
		p = p->next;

	if(p == NULL)
		return 0;
	else {
		aux = p;
		while(aux->next && aux->next->data->level->level == level) {
			aux = aux->next;
			if(!aux->data->commented)
				p = aux;
		}
	}

	if(p->data->nvalue == NULL)
		return 0;

	free(p->data->nvalue);
	p->data->nvalue = NULL;

	removeinstchange(conf, p->data);

	if(p->data->value == NULL) {
		freedesc(p->data->desc);
		free(p->data);
		p = var->data->instance;
		aux = p;
		while(p && (p->data->value == NULL || p->data->commented))
			p = p->next;

		if(p != NULL) {
			aux = p;
			while(aux->next) {
				aux = aux->next;
				if(!aux->data->commented &&
				    aux->data->value != NULL)
					p = aux;
			}
			var->data->nvalue = &p->data->nvalue;
			var->data->nlevel = p->data->level->level;
		}
	}

	return 0;
}

static void
addinstchange(struct configuration* conf, struct confinstdata* inst)
{
	struct confinst*	p;

	p = (struct confinst*) malloc(sizeof(struct confinst));

	p->data = inst;
	p->next = inst->level->instchanged;
	inst->level->instchanged = p;

	return;
}

static void
removeinstchange(struct configuration* conf, struct confinstdata* inst)
{
	struct confinst*	p;
	struct confinst*	ref;

	if(inst->level->instchanged) {
		if(inst->level->instchanged->data == inst) {
			ref = inst->level->instchanged;
			inst->level->instchanged =
			    inst->level->instchanged->next;
			free(ref);
		} else {
			p = inst->level->instchanged;
			while(p->next && p->next->data != inst)
				p = p->next;

			if(p->next != NULL) {
				ref = p->next;
				p->next = p->next->next;
				free(ref);
			}
		}
	}
	if(inst->level->instappended) {
		if(inst->level->instappended->data == inst) {
			ref = inst->level->instappended;
			inst->level->instappended =
			    inst->level->instappended->next;
			free(ref);
		} else {
			p = inst->level->instappended;
			while(p->next && p->next->data != inst)
				p = p->next;

			if(p->next != NULL) {
				ref = p->next;
				p->next = p->next->next;
				free(ref);
			}
		}
	}

	if(inst->level->instchanged == NULL &&
	    inst->level->instappended == NULL)
		inst->level->changed = 0;

	return;
}

struct confinst*
appendvarinstlevel(struct configuration* conf, int level, const char* name)
{
	struct confinst*	inst;
	struct confinst*	ref;
	struct confinstdata*	varinst;
	struct confvardata*	var;
	struct conflvl*		p;

	p = conf->levels;
	while(p && p->level != level)
		p = p->next;

	if(p == NULL)
		return NULL;

	varinst = (struct confinstdata*) malloc(sizeof(struct confinstdata));
	varinst->level = p;
	if(p->sections)
		varinst->section = p->sectail->data;
	else
		varinst->section = NULL;
	varinst->line = -1;
	varinst->value = NULL;
	varinst->nvalue = strdup("");
	varinst->pvalue = NULL;
	varinst->desc = (struct confdesc*) malloc(sizeof(struct confdesc));
	varinst->desc->nlines = 0;
	varinst->desc->line = NULL;
	varinst->desc->desc = NULL;
	varinst->commented = 0;

	var = getvariable(conf, p, name);
	conf->lastvar = var;
	varinst->data = var;

	inst = (struct confinst*) malloc(sizeof(struct confinst));
	inst->data = varinst;
	inst->next = NULL;

	if(var->instance == NULL) {
		var->instance = inst;
		var->insttail = inst;
	} else {
		ref = var->instance;
		while(ref && ref->data->level->level != level)
			ref = ref->next;
		while(ref && ref->next &&
		    ref->next->data->level->level == level)
			ref = ref->next;
		if(ref == NULL) {
			var->insttail->next = inst;
			var->insttail = inst;
		} else {
			inst->next = ref->next;
			ref->next = inst;
			if(inst->next == NULL)
				var->insttail->next = inst;
		}
	}

	if(var->desc == NULL)
		var->desc = varinst->desc;

	if(var->nlevel <= level) {
		var->nvalue = &varinst->nvalue;
		var->nlevel = level;
	}

	inst = (struct confinst*) malloc(sizeof(struct confinst));
	inst->data = varinst;
	inst->next = p->instappended;
	p->instappended = inst;

	p->changed = 1;

	return inst;
}

int
writeconf(struct configuration* conf)
{
	struct conflvl*		level;
	struct confinst*	inst;
	struct confinst*	ref;
	struct filecontent*	fc;
	int			fd;
	int			i;
	char*			p;
	int			writeerror, commiterror, rollbackerror;
	long			offset;
	off_t			filesize;

	writeerror = 0;
	commiterror = 0;
	rollbackerror = 0;

	level = conf->levels;
	while(level) {
		if(level->changed) {
			offset = 0;
			fc = level->fc;
			level->tmpfile = (char*)
			    malloc(strlen(level->filename)+6);
			strcpy(level->tmpfile, level->filename);
			strcat(level->tmpfile, ".XXXX");
			if((fd = mkstemp(level->tmpfile)) == -1) {
				writeerror = 1;
				break;
			}
			fchmod(fd, level->mode);
			fchown(fd, level->uid, level->gid);
			p = fc->data;
			for(i = 0; i < fc->nlines; i++)
				if(fc->lines[i]->instance &&
				    fc->lines[i]->instance->data->nvalue)
				    {
					inst = fc->lines[i]->instance;
					if(write(fd, p, fc->data - p +
					    inst->data->pvalue) == -1) {
						close(fd);
						writeerror = 1;
						break;
					}
					if(write(fd, "\"", 1) == -1 ||
					    write(fd, inst->data->nvalue,
					    strlen(inst->data->nvalue)) == -1
					    || write(fd, "\"", 1) == -1) {
						close(fd);
						writeerror = 1;
						break;
					}
					p = fc->data + inst->data->pvalue
					    + strlen(inst->data->value)
					    + (fc->data[inst->data->pvalue]
					    == '"' ||
					    fc->data[inst->data->pvalue]
					    == '\'' ? 2 : 0);
					offset += strlen(inst->data->nvalue) -
					    strlen(inst->data->value) + 2;
				}
			if(write(fd, p, fc->endp - p) == -1) {
				close(fd);
				writeerror = 1;
				break;
			}
			for(inst = level->instappended; inst;
			    inst = inst->next) {
				p = malloc(strlen(inst->data->data->name) +
				    strlen(inst->data->nvalue) + 4);
				strcpy(p, inst->data->data->name);
				strcat(p, "=\"");
				strcat(p, inst->data->nvalue);
				strcat(p, "\"\n");
				if(write(fd, p, strlen(p)) == -1) {
					close(fd);
					writeerror = 1;
					break;
				}
				offset += strlen(p);
				free(p);
			}

			if(close(fd) == -1) {
				writeerror = 1;
				break;
			}
		}
		level = level->next;
	}

	if(writeerror) {
		level = conf->levels;
		while(level) {
			if(level->tmpfile != NULL) {
				unlink(level->tmpfile);
				free(level->tmpfile);
				level->tmpfile = NULL;
			}
			level = level->next;
		}
	} else {
		level = conf->levels;
		while(level) {
			if(level->tmpfile != NULL) {
				level->oldfile = (char*)
			    	malloc(strlen(level->filename)+6);
				strcpy(level->oldfile, level->filename);
				strcat(level->oldfile, ".XXXX");
				if((fd = mkstemp(level->oldfile)) == -1) {
					free(level->oldfile);
					level->oldfile = NULL;
					commiterror = 1;
					break;
				}
				close(fd);
				unlink(level->oldfile);
				if(link(level->filename, level->oldfile)
				    == -1) {
					if(errno != EOPNOTSUPP &&
					    errno != ENOENT) {
						free(level->oldfile);
						level->oldfile = NULL;
						commiterror = 1;
						break;
					}
					if(errno == EOPNOTSUPP)
						if(rename(level->filename,
						    level->oldfile) == -1 &&
						    errno != ENOENT) {
							free(level->oldfile);
							level->oldfile = NULL;
							commiterror = 1;
							break;
						}
					free(level->oldfile);
					level->oldfile = NULL;
				}
				if(rename(level->tmpfile, level->filename)
				    == -1) {
					if(level->oldfile != NULL) {
						unlink(level->oldfile);
						free(level->oldfile);
						level->oldfile = NULL;
					}
					commiterror = 1;
					break;
				}
				free(level->tmpfile);
				level->tmpfile = NULL;
			}
			level = level->next;
		}
		if(commiterror) {
			level = conf->levels;
			while(level) {
				if(level->oldfile != NULL) {
					if(rename(level->oldfile,
					    level->filename) == -1)
						rollbackerror = 1;
					free(level->oldfile);
					level->oldfile = NULL;
				}
				if(level->tmpfile != NULL) {
					unlink(level->tmpfile);
					free(level->tmpfile);
					level->tmpfile = NULL;
				}
				level = level->next;
			}
		} else {
			level = conf->levels;
			while(level) {
				if(level->oldfile != NULL) {
					unlink(level->oldfile);
					free(level->oldfile);
					level->oldfile = NULL;
				}
				level = level->next;
			}
		}
	}

	if(rollbackerror)
		return -1;

	if(writeerror || commiterror)
		return 1;

	/* Everything went ok, so update all pointers and unmark
	 * changed's.
	 */

	level = conf->levels;
	while(level) {
		if(level->changed) {
			offset = 0;
			fc = level->fc;
			for(i = 0; i < fc->nlines; i++) {
				fc->lines[i]->line += offset;
				if(fc->lines[i]->instance) {
					inst = fc->lines[i]->instance;
					inst->data->pvalue += offset;
					if(fc->lines[i]->instance->data->nvalue
					    != NULL) {
						offset += 2 +
						    strlen(inst->data->nvalue) -
						    strlen(inst->data->value) -
						    (fc->data[
						    inst->data->pvalue] ==
						    '"' || fc->data[
						    inst->data->pvalue] ==
						    '\'' ? 2 : 0);
						free(inst->data->value);
						inst->data->value =
						    inst->data->nvalue;
						inst->data->nvalue = NULL;
					}
				}
			}
			for(inst = level->instappended; inst;) {
				if(fc->nlines % LINESINC == 0)
					fc->lines = (struct conflines**)
					    reallocf(fc->lines,
					    (fc->nlines / LINESINC + 1) *
					    LINESINC *
					    sizeof(struct conflines*));
				fc->lines[fc->nlines] = (struct conflines*)
				    malloc(sizeof(struct conflines));
				fc->lines[fc->nlines]->line = fc->endp -
				    fc->data + offset;
				inst->data->pvalue =
				    fc->lines[fc->nlines]->line +
				    strlen(inst->data->data->name) + 1;
				fc->lines[fc->nlines]->len =
				    strlen(inst->data->data->name) +
				    strlen(inst->data->nvalue) + 1;
				fc->lines[fc->nlines]->instance = inst;
				fc->nlines++;
				offset += strlen(inst->data->data->name) +
				    strlen(inst->data->nvalue) + 2;
				inst->data->value = inst->data->nvalue;
				inst->data->nvalue = NULL;
				ref = inst;
				inst = inst->next;
				ref->next = NULL;
			}
			level->instappended = NULL;
			free(fc->data);
			fd = open(level->filename, O_RDONLY);
			filesize = lseek(fd, (off_t)0, SEEK_END);
			lseek(fd, (off_t)0, SEEK_SET);
			fc->data = (char*) malloc((size_t)filesize+1);
			read(fd, fc->data, (size_t)filesize);
			fc->endp = fc->data + filesize;
			*fc->endp = 0;
			close(fd);
			inst = level->instchanged;
			while(inst) {
				ref = inst;
				inst = inst->next;
				free(ref);
			}
			level->instchanged = NULL;
			level->changed = 0;
		}
		level = level->next;
	}
	adjustevalue(conf);

	return 0;
}

struct confinst*
getfirstvarinst(struct confvar* var)
{
	if(var == NULL)
		return NULL;
	return var->data->instance;
}

struct confinst*
getnextvarinst(struct confinst* inst)
{
	if(inst == NULL)
		return NULL;
	return inst->next;
}

struct confvar*
varinst2var(struct confinst* inst)
{
	struct confvar* p;

	if(inst == NULL)
		return NULL;

	p = inst->data->level->conf->variables;
	while(p && CMPVAR(inst->data->level->conf,inst->data->data->name,
	    p->data->name))
		p = p->next;

	return p;
}

struct confvar*
varinst2varlevel(struct confinst* inst)
{
	struct confvar* p;

	if(inst == NULL)
		return NULL;

	p = inst->data->level->variables;
	while(p && CMPVAR(inst->data->level->conf,inst->data->data->name,
	    p->data->name))
		p = p->next;

	return p;
}

struct confvar*
varinst2varsec(struct confinst* inst)
{
	struct confvar* p;

	if(inst == NULL)
		return NULL;

	if(inst->data->section == NULL)
		return NULL;

	p = inst->data->section->variables;
	while(p && CMPVAR(inst->data->level->conf,inst->data->data->name,
	    p->data->name))
		p = p->next;

	return p;
}

long
varinstline(struct confinst* inst)
{
	if(inst == NULL)
		return NULL;
	return inst->data->line;
}

char*
browse(struct configuration* conf, int level, long line)
{
	struct conflvl*	p;
	char*		buffer;

	p = conf->levels;

	while(p && p->level != level)
		p = p->next;

	if(p == NULL)
		return NULL;

	if(p->fc->nlines < line)
		return NULL;

	buffer = (char*) malloc(p->fc->lines[line]->len + 1);
	strncpy(buffer, p->fc->data + p->fc->lines[line]->line,
	    p->fc->lines[line]->len);
	buffer[p->fc->lines[line]->len] = '\0';

	return buffer;
}

struct confinst*
browseinst(struct configuration* conf, int level, long line)
{
	struct conflvl*	p;

	p = conf->levels;

	while(p && p->level != level)
		p = p->next;

	if(p == NULL)
		return NULL;

	if(p->fc->nlines < line)
		return NULL;

	return p->fc->lines[line]->instance;
}

const char*
getvarinstvalue(struct confinst* inst)
{
	return inst->data->value;
}


static void
freedesc(struct confdesc* desc)
{
	long	i;
	for(i = 0; i < desc->nlines; i++)
		free(desc->line[i]);
	free(desc->desc);
}

static void adjustevalue(struct configuration* conf)
{
	struct confvar*		var;
	struct confinst*	inst;
	long			pvalue;

	for(var = conf->variables; var; var = var->next) {
		var->data->nlevel = -1;
		pvalue = 0;
		for(inst = var->data->instance; inst; inst = inst->next)
			if(!inst->data->commented &&
			    inst->data->level->level >= var->data->hlevel &&
			    inst->data->pvalue > pvalue) {
				var->data->hlevel = inst->data->level->level;
				var->data->evalue = &inst->data->value;
				var->data->nvalue = &inst->data->nvalue;
				pvalue = inst->data->pvalue;
			}
	}
}

int
unsetvarlevelbyname(struct configuration* conf, int level, const char* name)
{
	struct confvar* var;

	for(var = getfirstvarlevel(conf, level); var &&
	    CMPVAR(conf, var->data->name, name); var = var->next);

	if(var == NULL)
		return 1;

	return unsetvarlevel(var, level);
}

int
setvarlevelbyname(struct configuration* conf, int level, const char* name, const char* value)
{
	struct confvar* var;
	struct confinst* inst;

	for(var = getfirstvarlevel(conf, level); var &&
	    CMPVAR(conf, var->data->name, name); var = var->next);

	if(var == NULL) {
		inst = appendvarinstlevel(conf, level, name);
		if(inst == NULL)
			return 1;
		var = varinst2var(inst);
	}

	return setvarlevel(var, level, value);
}

struct confvar*
getvarbyname(struct configuration* conf, const char* name)
{
	struct confvar* var;

	for(var = conf->variables; var &&
	    CMPVAR(conf, var->data->name, name); var = var->next);

	return var;
}



