/*-
 * Copyright (c) 1999 Daniel Capo Sobral
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: libconf.h,v 1.14 1999/12/19 20:36:51 dcs Exp $
 */

#ifndef _LIBCONF_H_
#define _LIBCONF_H_

/* This .h was extracted from my work-in-progress, so don't bother
 * about it's ickness.
 *
 * NOTICE:
 *
 * None of the instance functions has been implemented. The groundwork
 * for that is already laid down though, and I'll happilly accept
 * patches. I want just want the lib of out the door first, though.
 */

/* Error codes returned */
#define SYNTAX_ERROR	(ELAST+1)	/* Non-fatal syntax error */
#define CANT_STAT	(ELAST+2)	/* Can't stat file */

struct configuration;	/* Configuration indexes */
struct confsec;		/* Section information */
struct confcomp;	/* Components of a section */
struct confvar;		/* Generic list of variables */
struct confinst;	/* Instance of a variable assignment */

/* createconf()  Initializes a new set of configuration files
 *
 * Returns a cookie used to identify what set of configuration files you
 * are using.
 *
 * A flag indicates whether the configuration is case sensitive or not.
 */
struct configuration* createconf(int);

/* readconf()  Read a configuration file
 *
 * The configuration file is stored as the next "level". The library
 * assumes this function will be called in the same order in which the
 * files get processed.
 */
int readconf(struct configuration*, const char*);

/* writeconf()  Write all configuration files with changes back
 * On success, returns 0. If an error happens, writeconf() will do all
 * possible to revert any changes already made. If it cannot roll-back
 * the changes, -1 is returned. If all changes are rolled-back, 1 is
 * returned.
 */
int writeconf(struct configuration*);

/* Lists are kept in such a way that the same functions to get the next
 * element can be used, no matter what ordering/subset you are
 * examining. These are defined by the first element you get. (Yes, that
 * means a LOT of information is kept around.)
 */

/* All variables, alphabetic order */
struct confvar* getfirstvar(struct configuration*);

/* All variables in a "level" (configuration file), alphabetic order */
struct confvar* getfirstvarlevel(struct configuration*, int);

/* All variables in a section, order of appearance */
struct confvar* getfirstvarsec(struct confsec*);

/* All instances of a variable, order of appearance */
struct confinst* getfirstvarinst(struct confvar*);

/* All sections, alphabetic order */
struct confsec* getfirstsection(struct configuration*);

/* All sections in a level, order of appearance */
struct confsec* getfirstsectionlevel(struct configuration*, int);

/* All subsections of a section, order of appearance */
struct confsec* getfirstsubsection(struct confsec*);

/* Get things by name */
struct confvar* getvarbyname(struct configuration*, const char*);
struct confvar* getvarlevelbyname(struct configuration*, int, const char*);
struct confvar* getvarsecbyname(struct configuration*, struct confsec*, const char*);
struct confvar* getvarseclevelbyname(struct configuration*, int, struct confsec*, const char*);
struct confsec* getsecbyname(struct configuration*, const char*);
struct confsec* getseclevelbyname(struct configuration*, int, const char*);
struct confsec* getsecsecbyname(struct configuration*, struct confsec*, const char*);
struct confsec* getsecseclevelbyname(struct configuration*, int, struct confsec*, const char*);

/* Get next of lists */
struct confsec* getnextsec(struct confsec*);
struct confvar* getnextvar(struct confvar*);
struct confinst* getnextvarinst(struct confinst*);

/* Information retrieval */

/* Description of a variable */
const char* getvardesc(struct confvar*);

/* Description of a variable at a given level */
const char* getvardesclevel(struct confvar*, int);

/* Description of a section */
const char* getsecdesc(struct confsec*);

/* Number of lines in the description of a section */
int getsecdescnlines(struct confsec*);

/* Line N of the description of a section */
const char* getsecdescline(struct confsec*, int);

/* Section's name */
const char* getsecname(struct confsec*);

/* Variable's name */
const char* getvarname(struct confvar*);

/* Variable's value */
const char* getvarvalue(struct confvar*);

/* Variable's previously defined highest level value */
const char* getvarovalue(struct confvar*);

/* Variable's value at some level */
const char* getvarvaluelevel(struct confvar*, int);

/* Instance's value */
const char* getvarinstvalue(struct confinst*);

/* Highest level at which a variable is defined */
int getvarlevel(struct confvar*);

/* Lowest level at which a configuration set was changed */
int getchangedlevel(struct configuration*);

/* Highest level at which the variable was changed */
int getvarchangelevel(struct confvar*);

/* Header level (top header = 0) */
int getseclevel(struct confsec*);

/* Configuration level the header belongs to */
int getsecconflevel(struct confsec*);

/* Was something changed? */
int confchanged(struct configuration*);

/* Was something changed at a specific level? */
int confchangedlevel(struct configuration*, int);

/* Was a variable changed? */
int varchanged(struct confvar*);

/* Which section does a variable belongs to? order of appearance */
struct confsec* getvarsec(struct confvar*);

/* Which section does a variable belongs to? order of appearance, by level */
struct confsec* getvarseclevel(struct confvar*);

/* Which section does a section belongs to? order of appearance, by section */
struct confsec* getsecsec(struct confsec*);

/* Variable a instance refers to, alphabetical order */
struct confvar* varinst2var(struct confinst*);

/* Variable a instance refers to, alphabetical order, configuration level */
struct confvar* varinst2varlevel(struct confinst*);

/* Variable a instance refers to, order of appearance, section level */
struct confvar* varinst2varsec(struct confinst*);

/* Is the variable commented-out at a certain level? */
int varcommented(struct confvar*, int);

/* Line number a instance was defined at */
long varinstline(struct confinst*);

/* Change information */
int setvarlevel(struct confvar*, int, const char*);
int setvarlevelbyname(struct configuration*, int, const char*, const char*);
int unsetvarlevel(struct confvar*, int);
int unsetvarlevelbyname(struct configuration*, int, const char*);

/* Append things */
struct confinst* appendvarinstlevel(struct configuration*, int, const char*);

/* Browse configuration files */
/* The functions returning strings will allocate memory for them, since
 * the file is stored internally as one big string, instead of one
 * string per line. Ie, free the memory after using it.
 */
long conflines(struct configuration*, int);
char* browse(struct configuration*, int, long);
struct confinst* browseinst(struct configuration*, int, long);

#endif /* !_LIBCONF_H_ */

